"""Grand Unified SNAP - Bringing together SNAP Connect + SNAP2MQTT under one API"""

import asyncore
import base64
import binascii
from datetime import timedelta
import functools
from itertools import count
import logging
import os
import random

from tornado.ioloop import IOLoop, PeriodicCallback, TimeoutError
from tornado.concurrent import Future
from tornado.gen import coroutine, with_timeout, sleep, Return
from tornado.queues import PriorityQueue, QueueEmpty

from timeit import default_timer as timer

# SNAP Connect
import apy
from snapconnect import snap
from snappy import SnapConstantsHelper as SnapConst
import snaplib

from snapconnect_futures import SnapConnectFutures

from snap_profile.profile import SCSerialParams, SCTcpParams

# MQTT Support
from snap2mqtt_client.client import BaseClient
from snap2mqtt_future_client.future_client import FutureClient

LOG = logging.getLogger(__name__)
QLOG = logging.getLogger("Gus_TxQ")
# This log defaults to INFO, if you *really* want to drink from the firehose,
# set "snapstraction.gus.txq"'s loglevel to DEBUG in your app.
QLOG.setLevel(logging.INFO)

TX_PERIOD_MS = 25
# Priority is a byte, 0x00 is highest priority, 0xFF is lowest. Default is somewhere in the middle.
DEFAULT_PRIORITY = 0x80

# METRICS TWEAKS
# EWMA == Exponentially Weighted Moving Average
# EWMA_ALPHA must be between 0 and 1 (degree of weighting decrease)
# 0 == Never include new data
# 1 == Never include old data
EWMA_ALPHA = 0.5

# THE METRICS
# Count of packets Rx'd
METRIC_PACKET_RX = 'packet_rx_count'
# Count of packets Tx'd
METRIC_PACKET_TX = 'packet_tx_count'
# Count of exceptions while Tx'ing
METRIC_TX_EXCEPTION = 'tx_exception_count'
# Total ms Gus has been running
METRIC_UPTIME = 'uptime_ms'
# Total time spent Tx'ing (includes waiting for callback responses, because that's time we don't spend Tx'ing other packets)
METRIC_TX_TOTAL_TIME = 'tx_time_ms'
# EWMA of time spent Tx'ing
METRIC_TX_MEAN_TIME = 'mean_tx_time_ms'
# Total time spent waiting for next Tx time (queue wait)
METRIC_TX_TOTAL_WAIT_TIME = 'tx_wait_time_ms'
# EWMA of time spent waiting for next Tx time (queue wait)
METRIC_TX_MEAN_WAIT_TIME = 'mean_tx_wait_time_ms'
# # of items in the Tx Queue
METRIC_TX_QUEUE_LENGTH = 'tx_queue_len'


class GusError(Exception):
    """Wrapper for Gus-specific exceptions"""
    pass


class NoResponseError(Exception):
    """Exception raised if there's no response"""
    pass


class Gus(object):
    """The Grand Unified SNAP interface"""

    def __new__(cls, profile, network, *args, **kwargs):
        """The Gus Factory Function"""
        LOG.debug("In a gus factory")
        if profile.kind in ('serial', 'tcp'):
            return object.__new__(SCFGus)  # To avoid recursing
        elif profile.kind == 'mqtt':
            return object.__new__(MQTTGus)  # To avoid recursing

        else:
            raise GusError("Unknown kind in profile %r" % profile)

    def __init__(self):
        self._rpc_callbacks = {}
        self._data_callbacks = []
        self._traceroute_callbacks = []

        self._metrics = {
            METRIC_PACKET_RX: 0,
            METRIC_PACKET_TX: 0,
            METRIC_TX_EXCEPTION: 0,
            METRIC_UPTIME: 0,
            METRIC_TX_TOTAL_TIME: 0,
            METRIC_TX_MEAN_TIME: None,  # Sentinel for EWMA metric
            METRIC_TX_TOTAL_WAIT_TIME: 0,
            METRIC_TX_MEAN_WAIT_TIME: None,  # Sentinel for EWMA metric
            METRIC_TX_QUEUE_LENGTH: 0,
        }

        self._start_time = 0

    def __repr__(self):
        return '%s(%s, bridge=%s)' % (self.__class__.__name__, self.address, self.bridge_address)

    def metrics(self):
        return self._metrics.copy()

    def _update_ewma_metric(self, key, value):
        if self._metrics[key] is None:
            self._metrics[key] = value
            return

        self._metrics[key] = int((EWMA_ALPHA * value) + ((1 - EWMA_ALPHA) * self._metrics[key]))

    def _update_uptime(self):
        self._metrics[METRIC_UPTIME] = int(1000 * (timer() - self._start_time))

    @property
    def bridge_address(self):
        """This should be the address of the bridge node.

        The bridge node is the node that gets us from serial to RF.
        """
        raise NotImplementedError

    @property
    def address(self):
        """This should be the 'self' address of the SNAP Interface.

        The self address is the address other nodes should send
        packets to if we want to receive them.
        """
        raise NotImplementedError

    def start(self):
        """Start Gus"""
        raise NotImplementedError

    def stop(self):
        """Stop Gus"""
        raise NotImplementedError

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def on_any_rpc(self, callback):
        """Register callback for any rpc"""
        cbs = self._rpc_callbacks.setdefault('*', [])

        if callback not in cbs:
            cbs.append(callback)

    def on_rpc(self, func_name, callback):
        """Register callback for a specific rpc"""
        func_cbs = self._rpc_callbacks.setdefault(func_name, [])

        if callback not in func_cbs:
            func_cbs.append(callback)

    def off_rpc(self, func_name, callback):
        """Unregister callback for a specific rpc"""
        func_cbs = self._rpc_callbacks.setdefault(func_name, [])

        if callback in func_cbs:
            func_cbs.remove(callback)

    def on_traceroute(self, callback):
        """Register callback for traceroutes"""
        if callback not in self._traceroute_callbacks:
            self._traceroute_callbacks.append(callback)

    def on_data(self, callback):
        """Register callback for data packets"""
        if callback not in self._data_callbacks:
            self._data_callbacks.append(callback)

    def _on_rpc_recv(self, src, func, *args):
        """Call registered RPC callbacks"""
        LOG.debug("Got an RPC: %s.%s%r", src, func, args)

        self._metrics[METRIC_PACKET_RX] += 1

        # Handle the function-specific callbacks
        for cb in self._rpc_callbacks.get(func, []):
            try:
                IOLoop.current().add_callback(cb, src, *args)
            except:  # pragma: no cover
                LOG.exception("RPC callback")

        # And the wildcard callbacks
        for cb in self._rpc_callbacks.get('*', []):
            try:
                IOLoop.current().add_callback(cb, src, func, *args)
            except:  # pragma: no cover
                LOG.exception("RPC Wildcard callback")

    def _on_data_recv(self, src, data):
        """Call registered Data callbacks"""

        self._metrics[METRIC_PACKET_RX] += 1

        for cb in self._data_callbacks:
            try:
                IOLoop.current().add_callback(cb, src, data)
            except:  # pragma: no cover
                LOG.exception("Data callback")

    def _on_traceroute_recv(self, src, rtt, hops):
        """Call registered Traceroute callbacks"""

        self._metrics[METRIC_PACKET_RX] += 1

        for cb in self._traceroute_callbacks:
            try:
                IOLoop.current().add_callback(cb, src, rtt, hops)
            except:  # pragma: no cover
                LOG.exception("Traceroute callback")

    def ucast_rpc(self, target, func, *args, **kwargs):
        """Unicast RPC"""
        raise NotImplementedError

    def mcast_rpc(self, func, *args, **kwargs):
        """Multicast RPC"""
        raise NotImplementedError

    def dmcast_rpc(self, targets, func, *args, **kwargs):
        """Directed Multicast RPC"""
        raise NotImplementedError

    def ucast_data(self, target, data, **kwargs):
        """Unicast Data"""
        raise NotImplementedError

    def mcast_data(self, data, **kwargs):
        """Multicast Data"""
        raise NotImplementedError

    def callback_ucast_rpc(self, target, func, *args, **kwargs):
        """Callback Unicast RPC"""
        raise NotImplementedError

    def callback_mcast_rpc(self, func, *args, **kwargs):
        """Callback Multicast RPC"""
        raise NotImplementedError

    def callback_dmcast_rpc(self, targets, func, *args, **kwargs):
        """Callback Directed Multicast RPC"""
        raise NotImplementedError

    def callback_ucast_data(self, target, data, **kwargs):
        """Callback Unicast Data"""
        raise NotImplementedError

    def callback_mcast_data(self, data, **kwargs):
        """Callback Multicast Data"""
        raise NotImplementedError

    def traceroute(self, target):
        """Traceroute"""
        raise NotImplementedError

    def run_sync(self, timeout, cmd, *args, **kwargs):
        """Run the command to completion"""
        cb = functools.partial(cmd, self, *args, **kwargs)
        try:
            return IOLoop.current().run_sync(cb, timeout=timeout)
        except TimeoutError:
            if timeout is None:
                LOG.debug("Timeout error for command that would've run forever, not an error.")
            else:
                LOG.error("Command timed out or aborted!")
                raise

    @staticmethod
    def from_existing_scf(scf_instance, bridge_addr=None):
        """Create a Gus instance from an existing SCF instance rather than creating a new one."""

        gus_from_existing_scf = object.__new__(SCFGus)
        gus_from_existing_scf.__init__(None, None, None, scf_instance=scf_instance, bridge_addr=bridge_addr)
        return gus_from_existing_scf


class SCFGus(Gus):
    """Wraps SCF in a Gus-like interface"""

    def __init__(self, profile, network, sc_license, *args, **kwargs):
        """Create a new Gus"""

        super(SCFGus, self).__init__()

        # Check to see if SCFGus is being built from an existing SCF Instance
        self.bridge_addr = kwargs.get('bridge_addr', None)
        self._scf = kwargs.get('scf_instance', None)
        if self._scf:
            self._sc = self._scf._snap
            self.from_existing_instance = True
        else:
            self.from_existing_instance = False

        # If using an existing SCF Instance, this setup will have already been done.
        if not self.from_existing_instance:
            self.profile = profile.as_sc_params()
            self.network = network.as_sc_params()
            self.license = sc_license

            if self.license:
                path = os.path.dirname(os.path.realpath(__file__))
                license_path = os.path.join(
                    path, os.path.expanduser(self.license))
                LOG.debug("SNAP Connect License: %s", license_path)
            else:
                license_path = None
                LOG.debug("SNAP Connect License: Licensed by Ethernet MAC")

            self.nvparam_path = os.path.expanduser("~/.snap/nvparams.dat")
            LOG.debug("SNAP Connect NVParam: %s", self.nvparam_path)

            # We want a clean slate
            try:
                os.remove(self.nvparam_path)
            except:
                # Might not exist yet.
                pass

            if self.profile.address:
                LOG.debug("SNAP Connect Address: %s", binascii.hexlify(self.profile.address))
            else:
                LOG.debug("SNAP Connect Address: By Ethernet MAC or First in License File")

            self._sc = snap.Snap(scheduler=apy.ioloop_scheduler.IOLoopScheduler.instance(),
                                 license_file=license_path,
                                 nvparams_file=self.nvparam_path,
                                 addr=self.profile.address if license_path else None,
                                 funcs={})

            LOG.debug("SNAP Connect Licensed MAC: %s", binascii.hexlify(self._sc.licensed_mac))

            self._scf = SnapConnectFutures(self._sc)

            self.connect_future = None
            self.disconnect_future = None

            self._sc.set_hook(snap.hooks.HOOK_SNAPCOM_OPENED, self._on_connect)
            self._sc.set_hook(snap.hooks.HOOK_SNAPCOM_CLOSED, self._on_disconnect)

            self._sc.save_nv_param(snap.NV_AES128_ENABLE_ID, self.network.encryption_type)
            self._sc.save_nv_param(snap.NV_AES128_KEY_ID, self.network.encryption_key)

            # Mesh Group Params, we're interested in all, and might forward all.
            self._sc.save_nv_param(snap.NV_GROUP_INTEREST_MASK_ID, 0xFFFF)
            self._sc.save_nv_param(snap.NV_GROUP_FORWARDING_MASK_ID, 0xFFFF)
            self._sc.save_nv_param(snap.NV_GROUP_SERIAL_FORWARDING_MASK_ID, 0xFFFF)

            # We're just talking to our bridge, don't delete routes
            self._sc.save_nv_param(snap.NV_MESH_ROUTE_AGE_MAX_TIMEOUT_ID, 0)

            # Borrowed more nvparams from SimplySnap...
            self._sc.save_nv_param(snap.NV_MESH_RREQ_WAIT_TIME_ID, 100)
            self._sc.save_nv_param(snap.NV_MESH_INITIAL_HOPLIMIT_ID, min(3, self.network.default_ttl))

            self._sc.save_nv_param(snap.NV_MESH_MAX_HOPLIMIT_ID, self.network.default_ttl)

            # We need to randomize our mesh routing packets
            initial_mesh_seq_num = random.randint(1, 255)
            self._sc.save_nv_param(snap.NV_MESH_SEQUENCE_NUMBER_ID, initial_mesh_seq_num)

            # Also have to reach into the bowels of SC because updating
            # the config doesn't actually update this value...
            snaplib.MeshCodec.MeshCodec.sequence_number = initial_mesh_seq_num
            LOG.debug("MESH SEQ NUMBER INITIALIZED: %d", initial_mesh_seq_num)

            # Make room for packet CRCs, and send out data CRCs,
            # but don't worry about them actually being there...
            # SNAP Connect is a more forgiving about this stuff...
            self._sc.save_nv_param(snap.NV_FEATURE_BITS_ID, 0x0500)
            snaplib.RpcCodec.appendCrc = True
            snaplib.RpcCodec.validateCrc = False

        # Hook into the IOLoop
        self._cb = PeriodicCallback(self._sc.poll_internals, 5)

        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_RPC, self._on_rpc_recv)
        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_RPC | SnapConst.PID_ENCRYPTION_MASK, self._on_rpc_recv)
        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_MULTICAST_RPC, self._on_rpc_recv)
        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_MULTICAST_RPC | SnapConst.PID_ENCRYPTION_MASK, self._on_rpc_recv)
        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_DIRECTED_MULTICAST_RPC, self._on_rpc_recv)
        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_DIRECTED_MULTICAST_RPC | SnapConst.PID_ENCRYPTION_MASK, self._on_rpc_recv)

        # Also register for Data Mode packets.
        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_DM_BCAST, self._on_data_recv)
        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_DM_BCAST | SnapConst.PID_ENCRYPTION_MASK, self._on_data_recv)
        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_DM_PTP, self._on_data_recv)
        self._sc.packet_dispatcher.register_pkt_type(SnapConst.PID_DM_PTP | SnapConst.PID_ENCRYPTION_MASK, self._on_data_recv)

        self._sc.set_hook(snap.hooks.HOOK_TRACEROUTE, self._on_traceroute_recv)

        # TX Queueing
        self._tx_queue = PriorityQueue()
        self._tx_period = TX_PERIOD_MS
        self._tx_count = count()

        self._start_time = timer()

    def _on_connect(self, connection_info, snap_address):
        """TCP connection callback"""
        LOG.info("Connection established: %s:%d, remote address is %s",
                 connection_info[0], connection_info[1], binascii.hexlify(snap_address))

        if self.connect_future is None:
            LOG.error("Unexpected TCP open call from %s:%d, remote address is %s",
                      connection_info[0], connection_info[1], binascii.hexlify(snap_address))
            return

        self.connect_future.set_result(binascii.hexlify(snap_address))

    def _on_disconnect(self, connection_info, snap_address):
        """TCP disconnect callback"""
        LOG.info("Connection Closed: %s:%d, remote address is %s",
                 connection_info[0], connection_info[1], binascii.hexlify(snap_address) if snap_address else 'None')

        if self.disconnect_future is None:
            LOG.error("Unexpected TCP close call from %s:%d", connection_info[0], connection_info[1])
            return

        self.disconnect_future.set_result(True)

    @property
    def bridge_address(self):
        return self.bridge_addr

    @property
    def address(self):
        return binascii.hexlify(self._sc.licensed_mac[-3:])

    def _on_rpc_recv(self, descriptor):
        super(SCFGus, self)._on_rpc_recv(binascii.hexlify(descriptor.srcAddr), descriptor.funcName, *descriptor.args)

    def _on_data_recv(self, descriptor):
        super(SCFGus, self)._on_data_recv(binascii.hexlify(descriptor.srcAddr), descriptor.data)

    def _on_traceroute_recv(self, src, rtt, hops):
        super(SCFGus, self)._on_traceroute_recv(binascii.hexlify(src), rtt, [{'addr': binascii.hexlify(hop[0]),
                                                                              'lq': hop[1]} for hop in hops])

    def _enqueue_tx(self, priority, tx_func):
        # Use insertion order as a tiebreaker between identical priorities.
        self._metrics[METRIC_TX_QUEUE_LENGTH] += 1
        self._tx_queue.put_nowait((priority, self._tx_count.next(), tx_func, timer()))

    def _schedule_tx_service(self):
        QLOG.debug("Scheduling TX Service")
        IOLoop.current().add_timeout(timedelta(milliseconds=self._tx_period), self._service_tx_queue)

    def _service_tx_queue(self):
        QLOG.debug("Servicing Tx Queue")

        self._update_uptime()

        try:
            priority, tx_count, tx_func, time_of_enqueue = self._tx_queue.get_nowait()
            self._metrics[METRIC_TX_QUEUE_LENGTH] -= 1

            @coroutine
            def try_exec():
                time_of_dispatch = timer()
                try:
                    # UPDATE METRICS
                    wait_ms = 1000 * (time_of_dispatch - time_of_enqueue)
                    QLOG.info("Executing enqueued TX after %d ms wait", wait_ms)
                    self._metrics[METRIC_TX_TOTAL_WAIT_TIME] += wait_ms
                    self._update_ewma_metric(METRIC_TX_MEAN_WAIT_TIME, wait_ms)

                    yield tx_func()
                except:
                    QLOG.exception("Couldn't process Tx")
                    self._metrics[METRIC_TX_EXCEPTION] += 1
                finally:
                    # UPDATE THE METRICS
                    time_of_completion = timer()
                    tx_time_ms = 1000 * (time_of_completion - time_of_dispatch)
                    QLOG.info("TX took %d ms", tx_time_ms)
                    self._metrics[METRIC_TX_TOTAL_TIME] += tx_time_ms
                    self._update_ewma_metric(METRIC_TX_MEAN_TIME, tx_time_ms)
                    self._metrics[METRIC_PACKET_TX] += 1
                    self._schedule_tx_service()

            IOLoop.current().add_callback(try_exec)
        except QueueEmpty:
            QLOG.debug("Nothing to Tx")
            self._schedule_tx_service()
        except:
            QLOG.exception("Servicing Tx Queue")
            self._schedule_tx_service()

    def ucast_rpc(self, target, func, *args, **kwargs):
        priority = kwargs.get('priority', DEFAULT_PRIORITY)
        LOG.info("Enqueuing UCAST RPC at Priority %02x for later execution: %s.%s%r", priority, target, func, args)

        # This future will resolve when the RPC is actually sent
        # (after being handled by the rate limiter)
        future = Future()

        @coroutine
        def f():
            LOG.info("Executing queued UCAST RPC: %s.%s%r", target, func, args)
            result = yield self._scf.rpc(target, func, *args)
            LOG.info("UCAST RPC Result: %r", result)
            future.set_result(result)

        self._enqueue_tx(priority, f)
        return future

    def mcast_rpc(self, func, *args, **kwargs):
        """Returns a future that will resolve when we actually Tx the mcast rpc"""
        priority = kwargs.get('priority', DEFAULT_PRIORITY)

        group = kwargs.get('group', self.network.default_group)
        ttl = kwargs.get('ttl', self.network.default_ttl)

        LOG.info("Enqueuing MCAST RPC at Priority %02x for later execution: %s.%s%r", priority, group, func, args)

        # This future will resolve when the RPC is actually sent
        # (after being handled by the rate limiter)
        future = Future()

        @coroutine
        def f():
            LOG.info("Executing queued MCAST RPC: %s.%s%r", group, func, args)
            result = yield self._scf.mcast_rpc(func, *args, group=group, ttl=ttl)
            LOG.info("MCAST RPC Result: %r", result)
            future.set_result(result)

        self._enqueue_tx(priority, f)
        return future

    def dmcast_rpc(self, targets, func, *args, **kwargs):
        """Returns a future that will resolve when we actually Tx the dmcast rpc"""
        priority = kwargs.get('priority', DEFAULT_PRIORITY)

        group = kwargs.get('group', self.network.default_group)
        ttl = kwargs.get('ttl', self.network.default_ttl)
        delay = kwargs.get('delay', kwargs.get('delay_factor', self.network.default_delay))

        LOG.info("Enqueuing DMCAST RPC at Priority %02x for later execution: %s.%s%r", priority, group, func, args)

        # This future will resolve when the RPC is actually sent
        # (after being handled by the rate limiter)
        future = Future()

        @coroutine
        def f():
            LOG.info("Executing queued DMCAST RPC: %s.%s%r", group, func, args)
            result = yield self._scf.dmcast_rpc(targets, func, *args, group=group, ttl=ttl, delay=delay)
            LOG.info("DMCAST RPC Result: %r", result)
            future.set_result(result)

        self._enqueue_tx(priority, f)
        return future

    @coroutine
    def _do_with_retries(self, func, retries, initial_priority, final_future):
        priority = initial_priority
        for _ in xrange(retries + 1):

                future = func(priority=priority)
                try:
                    result = yield future
                    if result is None:
                        # No Response, we want to retry, let our retries have a slightly higher priority
                        priority = max(0, priority - 0x04)
                        continue
                    else:
                        final_future.set_result(result)
                        return
                except Exception as e:
                    # Weird, just pass the exception up and stop
                    final_future.set_exception(e)
                    return

        # If we get here, we've exhausted our retries
        final_future.set_exception(NoResponseError())

    def _enqueue_cb_ucast_rpc(self, target, func, *args, **kwargs):
        timeout = kwargs.get("timeout", self.network.default_cb_timeout)
        retries = kwargs.get("retries", self.network.default_cb_retries)
        result_type = kwargs.get("result_type", None)
        callback_name = kwargs.get("callback_name", None)
        validate_func = kwargs.get("validate_func", None)

        priority = kwargs.get('priority', DEFAULT_PRIORITY)

        future = Future()

        @coroutine
        def f():
            LOG.info("Executing queued CALLBACK UCAST RPC: %s.%s%r", target, func, args)
            result = yield self._scf.callback_rpc(target, func, *args, timeout=timeout, retries=retries,
                                                  result_type=result_type, callback_name=callback_name,
                                                  validate_func=validate_func)
            LOG.info("CALLBACK UCAST RPC Result: %r", result)
            future.set_result(result)

        LOG.info("Enqueuing CALLBACK UCAST RPC at Priority %02x for later execution: %s.%s%r", priority, target, func, args)

        self._enqueue_tx(priority, f)
        return future

    def callback_ucast_rpc(self, target, func, *args, **kwargs):
        timeout = kwargs.get("timeout", self.network.default_cb_timeout)
        retries = kwargs.get("retries", self.network.default_cb_retries)
        result_type = kwargs.get("result_type", None)
        callback_name = kwargs.get("callback_name", None)
        validate_func = kwargs.get("validate_func", None)

        priority = kwargs.get('priority', DEFAULT_PRIORITY)

        future = Future()

        func = functools.partial(self._enqueue_cb_ucast_rpc, target, func, *args, timeout=timeout, retries=0,
                                 result_type=result_type, callback_name=callback_name, validate_func=validate_func,
                                 priority=priority)

        IOLoop.current().add_callback(self._do_with_retries, func, retries, priority, future)

        return future

    def _enqueue_cb_mcast_rpc(self, targets, func, *args, **kwargs):
        timeout = kwargs.get("timeout", self.network.default_cb_timeout)
        retries = kwargs.get("retries", self.network.default_cb_retries)
        result_type = kwargs.get("result_type", None)
        callback_name = kwargs.get("callback_name", None)
        ttl = kwargs.get("ttl", self.network.default_ttl)
        group = kwargs.get("group", self.network.default_group)
        validate_func = kwargs.get("validate_func", None)

        priority = kwargs.get('priority', DEFAULT_PRIORITY)

        LOG.info("Enqueuing CALLBACK MCAST RPC at Priority %02x for later execution: %s.%s%r", priority, targets, func, args)

        future = Future()

        @coroutine
        def f():
            LOG.info("Executing queued CALLBACK MCAST RPC: %s.%s%r", targets, func, args)
            result = yield self._scf.callback_mcast_rpc(targets, func, *args, timeout=timeout, retries=retries,
                                                        result_type=result_type, callback_name=callback_name,
                                                        ttl=ttl, group=group, validate_func=validate_func)
            LOG.info("CALLBACK MCAST RPC Result: %r", result)
            if result is None:
                future.set_exception(NoResponseError())
            else:
                future.set_result(result)

        self._enqueue_tx(priority, f)
        return future

    def callback_mcast_rpc(self, targets, func, *args, **kwargs):
        timeout = kwargs.get("timeout", self.network.default_cb_timeout)
        retries = kwargs.get("retries", self.network.default_cb_retries)
        result_type = kwargs.get("result_type", None)
        callback_name = kwargs.get("callback_name", None)
        ttl = kwargs.get("ttl", self.network.default_ttl)
        group = kwargs.get("group", self.network.default_group)
        validate_func = kwargs.get("validate_func", None)

        priority = kwargs.get('priority', DEFAULT_PRIORITY)

        future = Future()

        func = functools.partial(self._enqueue_cb_mcast_rpc, targets, func, *args,
                                 timeout=timeout, retries=retries,
                                 result_type=result_type, callback_name=callback_name,
                                 ttl=ttl, group=group, priority=priority,
                                 validate_func=validate_func)

        IOLoop.current().add_callback(self._do_with_retries, func, retries, priority, future)

        return future

    def _enqueue_cb_dmcast_rpc(self, targets, func, *args, **kwargs):
        timeout = kwargs.get("timeout", self.network.default_cb_timeout)
        retries = kwargs.get("retries", self.network.default_cb_retries)
        result_type = kwargs.get("result_type", None)
        callback_name = kwargs.get("callback_name", None)
        ttl = kwargs.get("ttl", self.network.default_ttl)
        group = kwargs.get("group", self.network.default_group)
        delay = kwargs.get("delay", self.network.default_delay)
        node_limit = kwargs.get("node_limit", None)
        chunk_results = kwargs.get("chunk_results", False)
        previous_chunk = kwargs.get("previous_chunk", None)
        validate_func = kwargs.get("validate_func", None)

        priority = kwargs.get('priority', DEFAULT_PRIORITY)

        LOG.info("Enqueuing CALLBACK DMCAST RPC at Priority %02x for later execution: %s.%s%r", priority, targets, func, args)

        future = Future()

        @coroutine
        def f():
            LOG.info("Executing queued CALLBACK DMCAST RPC: %s.%s%r", targets, func, args)
            result = yield self._scf.callback_dmcast_rpc(targets, func, *args, timeout=timeout, retries=retries,
                                                         result_type=result_type, callback_name=callback_name,
                                                         dmcast_delay=delay, node_limit=node_limit, ttl=ttl,
                                                         group=group, expect_seq=False, chunk_results=chunk_results,
                                                         previous_chunk=previous_chunk, validate_func=validate_func)
            LOG.info("CALLBACK DMCAST RPC Result: %r", result)
            if result is None:
                future.set_exception(NoResponseError())
            else:
                future.set_result(result)

        self._enqueue_tx(priority, f)
        return future

    def callback_dmcast_rpc(self, targets, func, *args, **kwargs):
        timeout = kwargs.get("timeout", self.network.default_cb_timeout)
        retries = kwargs.get("retries", self.network.default_cb_retries)
        result_type = kwargs.get("result_type", None)
        callback_name = kwargs.get("callback_name", None)
        ttl = kwargs.get("ttl", self.network.default_ttl)
        group = kwargs.get("group", self.network.default_group)
        delay = kwargs.get("delay", self.network.default_delay)
        node_limit = kwargs.get("node_limit", None)
        validate_func = kwargs.get("validate_func", None)

        priority = kwargs.get('priority', DEFAULT_PRIORITY)

        LOG.info("Enqueuing CALLBACK DMCAST RPC at Priority %02x for later execution: %s.%s%r", priority, targets, func, args)

        future = Future()

        @coroutine
        def f(dmcast, priority=priority, last_result=None):
            LOG.info("Executing queued CALLBACK DMCAST RPC: %s.%s%r", targets, func, args)
            result = yield dmcast()
            LOG.info("CALLBACK DMCAST RPC Result: %r", result)
            if result:
                priority = max(0, priority - 0x04)
                LOG.info("Enqueuing RETRY CALLBACK DMCAST RPC at Priority %02x for later execution: %s.%s%r", priority, targets, func, args)
                self._enqueue_tx(priority, functools.partial(f, result.retry, priority=priority, last_result=result))
                return
            else:
                # Must've finished
                if last_result is None or not last_result.successful:
                    future.set_exception(NoResponseError())
                else:
                    future.set_result(last_result)

        def initial_dmcast():
            return self._scf.callback_dmcast_rpc(targets, func, *args, timeout=timeout, retries=retries,
                                                 result_type=result_type, callback_name=callback_name,
                                                 dmcast_delay=delay, node_limit=node_limit, ttl=ttl,
                                                 group=group, expect_seq=False, chunk_results=True,
                                                 validate_func=validate_func)

        self._enqueue_tx(priority, functools.partial(f, initial_dmcast, priority=priority))
        return future

    def ucast_data(self, target, data, **kwargs):
        # Support base64 data
        if kwargs.get('base64', False):
            data = base64.b64decode(data)

        priority = kwargs.get('priority', DEFAULT_PRIORITY)

        LOG.info("Enqueuing UCAST DATA at Priority %02x for later execution: %s.%r", priority, target, data)

        future = Future()
        rpc_future = Future()

        @coroutine
        def f():
            LOG.info("Executing queued UCAST DATA: %s.%r", target, data)
            rpc_id = self._scf._snap.data_mode(binascii.unhexlify(target), data)

            if self._scf._immediately_resolve_rpcs:
                rpc_future.set_result(True)
                self._scf._message_sent()
            else:
                self._scf._rpc_futures[rpc_id] = rpc_future

            result = yield rpc_future

            future.set_result(result)

        self._enqueue_tx(priority, f)
        return future

    def mcast_data(self, data, **kwargs):
        group = kwargs.get('group', self.network.default_group) or self.network.default_group
        ttl = kwargs.get('ttl', self.network.default_ttl) or self.network.default_ttl

        # Support base64 data
        if kwargs.get('base64', False):
            data = base64.b64decode(data)

        priority = kwargs.get('priority', DEFAULT_PRIORITY)

        LOG.info("Enqueuing MCAST DATA at Priority %02x for later execution: %s.%r", priority, group, data)

        future = Future()
        rpc_future = Future()

        @coroutine
        def f():
            LOG.info("Executing queued MCAST DATA: %s.%r", group, data)
            rpc_id = self._scf._snap.mcast_data_mode(group, ttl, data)

            if self._scf._immediately_resolve_rpcs:
                rpc_future.set_result(True)
                self._scf._message_sent()
            else:
                self._scf._rpc_futures[rpc_id] = rpc_future

            result = yield rpc_future

            future.set_result(result)

        self._enqueue_tx(priority, f)
        return future

    def traceroute(self, target):
        self._scf._snap.traceroute(binascii.unhexlify(target))

    @coroutine
    def _init_for_tcp(self):
        """Initialize for SNAPconnect TCP connection"""

        if self.connect_future and self.connect_future.running():
            raise Exception("Already connecting")

        self.connect_future = Future()

        # SNAP Connect TCP requires asyncore be polled as well.
        self._asyncpoll = PeriodicCallback(asyncore.poll, 5).start()

        def client_auth(realm):
            """SNAPconnect TCP client authentication"""
            return (self.profile.username, self.profile.password)

        self._scf._snap.connect_tcp(self.profile.host,
                                    auth_info=client_auth,
                                    port=self.profile.port,
                                    secure=self.profile.ssl)

        bridge_addr = yield with_timeout(timedelta(seconds=10), self.connect_future)

        if bridge_addr is None:
            raise Exception("Unable to open connection!")

        self.bridge_addr = bridge_addr

    @coroutine
    def _init_for_serial(self):
        """Initialize for SNAPconnect serial connection"""
        LOG.info("Opening serial type: %r, port %r", self.profile.type, self.profile.port)

        bridge_addr = yield self._scf.open_serial(self.profile.type, self.profile.port)

        if bridge_addr is None:
            raise Exception("Unable to open serial bridge!")

        self.bridge_addr = bridge_addr

    def start(self):
        LOG.debug("In start")

        self._cb.start()
        self._schedule_tx_service()

        LOG.debug("callback started")

        # Check to see if SFCGus was built off an existing SFC Instance
        if not self.from_existing_instance:
            if isinstance(self.profile, SCSerialParams):
                try:
                    return IOLoop.current().run_sync(self._init_for_serial, timeout=30)
                except TimeoutError:
                    LOG.error("Timeout or abort starting serial!")
            elif isinstance(self.profile, SCTcpParams):
                try:
                    return IOLoop.current().run_sync(self._init_for_tcp, timeout=30)
                except TimeoutError:
                    LOG.error("Timeout or abort starting TCP!")
            else:
                raise GusError("%s is not a valid profile" % self.profile)

    @coroutine
    def _close_for_serial(self):
        """Shut down SNAPconnect serial connection"""
        # Give SNAP Connect a little bit longer to get packets out
        yield sleep(0.5)

        yield self._scf.close_serial(self.profile.type, self.profile.port)

    @coroutine
    def _close_for_tcp(self):
        """Shut down SNAPconnect TCP connection"""
        # Give SNAP Connect a little bit longer to get packets out
        yield sleep(0.5)

        if self.disconnect_future and self.disconnect_future.running():
            raise Exception("Already disconnecting")

        self.disconnect_future = Future()

        self._scf._snap.disconnect_tcp(self.profile.host, port=self.profile.port)

        yield with_timeout(timedelta(seconds=10), self.disconnect_future)

    def stop(self):
        # Close things down
        # Check to see if SFCGus was built off an existing SFC Instance
        if not self.from_existing_instance:
            if isinstance(self.profile, SCSerialParams):
                try:
                    return IOLoop.current().run_sync(self._close_for_serial, timeout=30)
                except TimeoutError:
                    LOG.error("Timeout or abort closing serial!")
            elif isinstance(self.profile, SCTcpParams):
                try:
                    return IOLoop.current().run_sync(self._close_for_tcp, timeout=30)
                except TimeoutError:
                    LOG.error("Timeout or abort closing TCP!")
            else:
                raise GusError("%s is not a valid profile" % self.profile)

        self._cb.stop()

        # Clean up after yourself
        try:
            os.remove(self.nvparam_path)
        except:
            pass


class MQTTGus(Gus):
    """Wraps SNAP2MQTTFutures in a Gus-like interface"""

    def __init__(self, profile, network, *args, **kwargs):
        """Create a new Gus"""

        super(MQTTGus, self).__init__()

        LOG.debug("New MQTTGus with profile: %r", profile)

        self.addr = profile.address

        self.profile = profile
        self.network = network

        self._base_client = BaseClient(profile.address, mqtt_host=profile.host)
        self._client = FutureClient(self._base_client)

        # Alias commands up
        self.ucast_rpc = self._client.uRPC

    @property
    def bridge_address(self):
        return self._base_client.bridge_address

    @property
    def address(self):
        return self.addr

    @coroutine
    def _init_for_mqtt(self):
        """Connect the MQTT client"""

        self._base_client.start()

        connect_future = Future()
        address_future = Future()

        def on_connect(client, userdata, flags, rc):
            """MQTT connect callback"""
            LOG.info("Connected to MQTT with result code %s", rc)
            connect_future.set_result(True)

        def on_address_change(address):
            LOG.info("MQTT Bridge Address: %r", address)
            try:
                address_future.set_result(True)
            except:
                LOG.exception("Got multiple address changed notifications")

        self._base_client.on_connect(on_connect)
        self._base_client.on_address_change(on_address_change)

        self._base_client.subscribe('rpc/#', self._on_rpc_recv)
        self._base_client.subscribe('data/#', self._on_data_recv)
        self._base_client.subscribe('traceroute/#', self._on_traceroute_recv)

        self._uptime_update = PeriodicCallback(self._update_uptime, 1000)

        yield with_timeout(timedelta(seconds=10), connect_future)
        yield with_timeout(timedelta(seconds=10), address_future)

    def start(self):
        try:
            self._start_time = timer()
            self._uptime_update.start()
            return IOLoop.current().run_sync(self._init_for_mqtt, timeout=30)
        except TimeoutError:
            LOG.error("Timeout or abort starting MQTT!")

    def stop(self):
        """Stop the MQTT client"""
        self._uptime_update.stop()
        self._base_client.stop()

    def _on_rpc_recv(self, pkt):
        super(MQTTGus, self)._on_rpc_recv(pkt['src'], pkt['func'], *pkt['args'])

    def _on_data_recv(self, pkt):
        super(MQTTGus, self)._on_data_recv(pkt['src'], pkt['data'])

    def _on_traceroute_recv(self, pkt):
        super(MQTTGus, self)._on_traceroute_recv(pkt['src'], pkt['rtt'], pkt['hops'])

    def mcast_rpc(self, funcname, *args, **kwargs):
        group = kwargs.get('group', self.network.default_group)
        ttl = kwargs.get('ttl', self.network.default_ttl)

        return self._client.mRPC(group, ttl, funcname, *args)

    def dmcast_rpc(self, addrs, funcname, *args, **kwargs):

        group = kwargs.get('group', self.network.default_group)
        ttl = kwargs.get('ttl', self.network.default_ttl)
        delay = kwargs.get('delay', self.network.default_delay)

        return self._client.dmRPC(addrs, group, ttl, delay, funcname, *args)

    def ucast_data(self, addr, data, **kwargs):
        # Support base64 data
        if kwargs.get('base64', False):
            data = base64.b64decode(data)

        return self._client.uData(addr, data)

    def mcast_data(self, data, **kwargs):

        group = kwargs.get('group', self.network.default_group)
        ttl = kwargs.get('ttl', self.network.default_ttl)

        # Support base64 data
        if kwargs.get('base64', False):
            data = base64.b64decode(data)

        return self._client.mData(group, ttl, data)

    @coroutine
    def callback_ucast_rpc(self, addr, funcname, *args, **kwargs):
        timeout = kwargs.get("timeout", self.network.default_cb_timeout)
        retries = kwargs.get("retries", self.network.default_cb_retries)
        result_type = kwargs.get("result_type", None)
        callback_name = kwargs.get("callback_name", None)

        result = yield self._client.callback_rpc(addr, funcname, args, timeout=timeout, retries=retries,
                                                 result_type=result_type, callback_name=callback_name)
        if result is None:
            raise NoResponseError()

        raise Return(result)

    @coroutine
    def callback_mcast_rpc(self, addrs, funcname, *args, **kwargs):
        timeout = kwargs.get("timeout", self.network.default_cb_timeout)
        retries = kwargs.get("retries", self.network.default_cb_retries)
        result_type = kwargs.get("result_type", None)
        callback_name = kwargs.get("callback_name", None)
        group = kwargs.get("group", self.network.default_group)
        ttl = kwargs.get("ttl", self.network.default_ttl)

        result = yield self._client.callback_mcast_rpc(addrs, funcname, args,
                                                       timeout=timeout, retries=retries,
                                                       result_type=result_type,
                                                       callback_name=callback_name,
                                                       group=group, ttl=ttl)
        if result is None:
            raise NoResponseError()

        raise Return(result)

    @coroutine
    def callback_dmcast_rpc(self, addrs, funcname, *args, **kwargs):
        timeout = kwargs.get("timeout", self.network.default_cb_timeout)
        retries = kwargs.get("retries", self.network.default_cb_retries)
        result_type = kwargs.get("result_type", None)
        callback_name = kwargs.get("callback_name", None)
        group = kwargs.get("group", self.network.default_group)
        ttl = kwargs.get("ttl", self.network.default_ttl)
        dmcast_delay = kwargs.get("dmcast_delay", self.network.default_delay)
        node_limit = kwargs.get("node_limit", None)
        chunk_results = kwargs.get("chunk_results", False)
        previous_chunk = kwargs.get("previous_chunk", None)

        result = yield self._client.callback_dmcast(addrs, funcname, args,
                                                    timeout=timeout, retries=retries,
                                                    result_type=result_type,
                                                    callback_name=callback_name,
                                                    dmcast_delay=dmcast_delay,
                                                    node_limit=node_limit, ttl=ttl,
                                                    chunk_results=chunk_results,
                                                    previous_chunk=previous_chunk,
                                                    group=group)

        if result is None:
            raise NoResponseError()

        raise Return(result)

    def traceroute(self, target):
        return self._client.traceroute(target)
