"""SNAPstraction Utility Functions"""
# Future SNAP Connect Context Manager
import logging

import re

import os
import sys

from distutils.sysconfig import get_python_lib

log = logging.getLogger(__name__)


ADDR_RE = re.compile(r'^[0-9A-Fa-f]{6}$')


CRYPTO_LOOKUP = {0: "No",
                 1: "AES-128",
                 2: "Basic"}


def first_or_none(result):
    if isinstance(result, list):
        result = tuple(result)

    if result is not None and isinstance(result, tuple):
        log.debug("Got a tuple result: %r", result)
        return result[0]
    else:
        log.debug("Got a non-tuple result: %r", result)
        return result


def parse_value_for_snap(string):
    """Try to parse a value for use with SNAP.

    Tries to parse as integer first: '123' => 123, '0x20' => 32

    "" and '' are both treated as empty strings.
    Outermost quotes are removed: 'abc' => abc.
    Boolean-ish things are booleans: True/true => True, False/false => False
    None-ish strings are None: None/none => None

    (To use boolean/none-ish strings as literals, encase in quotes: "True")

    Strings are decoded using Python's "string_escape" codec:
    "\xde\xca\xfb\xad" is a 4-byte string, and so is 'ab\xcd\xef'.

    """
    if string is None:
        return None

    try:
        # Maybe it's an integer?
        return int(string, 0)
    except:
        # Is it an empty string?
        if string == "''" or string == '""' or string == "":
            return ""

        # Is it a single-quoted string?
        if string[0] == "'" and string[-1] == "'" and len(string) > 2:
            return string[1:-1].decode('string_escape')

        # Is it a double-quoted string?
        if string[0] == '"' and string[-1] == '"' and len(string) > 2:
            return string[1:-1].decode('string_escape')

        # Is it boolean-y
        if string.lower() == "true":
            return True
        if string.lower() == "false":
            return False

        # Is it None?
        if string.lower() == "none":
            return None
    return string.decode('string_escape')


def validate_snap_address(addr):
    if not isinstance(addr, basestring):
        return None

    address = addr.lower()

    if address == 'bridge' or (ADDR_RE.match(address) is not None):
        return address
    else:
        return None


def clean_up(t, bridge):
    """If the target is 'bridge', replace it with the bridge address."""
    t = validate_snap_address(t)
    if t == 'bridge':
        t = bridge

    return t


def fix_up_target(func):
    """Wrap the func with "clean up targets first"

    Args:
        func (function): function to wrap with "clean up/replace target"

    Returns:
        Wrapped function.

    """

    def fixed(gus, target, *args, **kwargs):
        # Clean up iterable of targets, too.
        if isinstance(target, basestring):
            target = clean_up(target, gus.bridge_address)
        else:
            try:
                target = iter(target)
            except:
                raise Exception("%r was not a valid target" % target)
            else:
                target = [clean_up(t, gus.bridge_address) for t in target]

        return func(gus, target, *args, **kwargs)

    return fixed


def executing_in_virtualenv():
    """Check if function is being executed in a virtual environment.

    Returns:
        bool, if the function is executed in a virtualenv it will return True, False otherwise
    """
    return 'VIRTUAL_ENV' in os.environ


def get_virtualenv_path():
    """Get the path to the virtual environment Python library

    Returns:
        str, the path to the virtual environment Python library
    """
    return os.path.abspath(get_python_lib(prefix=os.environ['VIRTUAL_ENV']))


def ensure_path_contains_python_lib():
    """Ensure that the appropriate Python library path(s) are included in sys.path

    If the sys.path does not include the "installed" Python library path for
    SNAPstraction, add that library path to the end of sys.path. Furthermore,
    if we are currently running in a virtual environment, add that virtual environment's
    library path to the front of sys.path (so that its entries will take precedence
    over other modules found later in sys.path).

    On return, sys.path will contain at most two new entries.
    """
    python_lib = get_python_lib()
    if python_lib not in sys.path:
        sys.path.append(python_lib)

    if executing_in_virtualenv():
        virtualenv_path = get_virtualenv_path()
        if virtualenv_path not in sys.path:
            sys.path.insert(1, virtualenv_path)


class OutputFilter(object):
    """Context Manager for filtering extraneous log messages during some repetitive tasks.  All log messages from the
      given source will be recast as DEBUG.  Automatically attaches and removes itself as a filter to the given logger
      if used as a Context Manager.

    Args:
        name (str): Name of logger to be filtered.

    Raises:
        KeyError: Logger provided is not an existing logger.
    """
    def __init__(self, name):
        if name not in logging.Logger.manager.loggerDict:
            raise KeyError('%s is not an existing logger.' % name)

        self.name = name
        self.log = logging.getLogger(self.name)

    def __enter__(self):
        self.log.addFilter(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.log.removeFilter(self)

        if exc_type is not None:
            return False

    def filter(self, record):
        if not record.levelno == logging.DEBUG:
            self.log.debug(record.getMessage())
            return False
        return True
